@@ -14,12 +14,13 @@ import requests
 import subprocess
 
 from distutils.spawn import find_executable
+from datetime import datetime
 from bs4 import BeautifulSoup
 from requests import HTTPError
-from six import print_ as print
 from six.moves import urllib_parse, input
 
 from figcli.config import VERSION
+from figcli.io.output import Output
 from figcli.svcs.mfa.u2f import *
 
 from figcli.io.input import Input
@@ -34,7 +35,7 @@ class ExpectedGoogleException(Exception):
 
 
 class Google:
-    def __init__(self, config, save_failure, defaults: CLIDefaults, secrets_mgr: SecretsManager):
+    def __init__(self, config, save_failure, defaults: CLIDefaults, secrets_mgr: SecretsManager, save_flow=False):
         """The Google object holds authentication state
         for a given session. You need to supply:
         username: FQDN Google username, eg first.last@example.com
@@ -51,8 +52,14 @@ class Google:
         self.save_failure = save_failure
         self.session_state = None
         self._defaults = defaults
+        self.save_flow = save_flow
         self._secrets_mgr = secrets_mgr
+        self._out = Output(defaults.colors_enabled)
 
+        if save_flow:
+            self.save_flow_dict = {}
+            self.save_flow_dir = "aws-google-auth-" + datetime.now().strftime('%Y-%m-%dT%H%M%S')
+            os.makedirs(self.save_flow_dir, exist_ok=True)
     @property
     def login_url(self):
         return self.base_url + "/o/saml2/initsso?idpid={}&spid={}&forceauthn=false".format(
@@ -89,9 +96,35 @@ class Google:
 
         return sess
 
-    def post(self, url, data=None, json=None):
+    def _save_file_name(self, url):
+        filename = url.split('://')[1].split('?')[0].replace("accounts.google", "ac.go").replace("/", "~")
+        file_idx = self.save_flow_dict.get(filename, 1)
+        self.save_flow_dict[filename] = file_idx + 1
+        return filename + "_" + str(file_idx)
+
+    def _save_request(self, url, method='GET', data=None, json_data=None):
+        if self.save_flow:
+            filename = self._save_file_name(url) + "_" + method + ".req"
+            with open(os.path.join(self.save_flow_dir, filename), 'w', encoding='utf-8') as out:
+                try:
+                    out.write("params=" + url.split('?')[1])
+                except IndexError:
+                    out.write("params=None")
+                out.write(("\ndata: " + json.dumps(data, indent=2)).replace(self.config.password, '<PASSWORD>'))
+                out.write(("\njson: " + json.dumps(json_data, indent=2)).replace(self.config.password, '<PASSWORD>'))
+
+    def _save_response(self, url, response):
+        if self.save_flow:
+            filename = self._save_file_name(url) + ".html"
+            with open(os.path.join(self.save_flow_dir, filename), 'w', encoding='utf-8') as out:
+                out.write(response.text)
+
+    def post(self, url, data=None, json_data=None):
         try:
-            response = self.check_for_failure(self.session.post(url, data=data, json=json))
+            self._save_request(url, method='POST', data=data, json_data=json_data)
+            response = self.check_for_failure(self.session.post(url, data=data, json=json_data))
+            self._save_response(url, response)
+
         except requests.exceptions.ConnectionError as e:
             logging.exception(
                 'There was a connection error, check your network settings.', e)
@@ -108,7 +141,10 @@ class Google:
 
     def get(self, url):
         try:
+            self._save_request(url)
             response = self.check_for_failure(self.session.get(url))
+            self._save_response(url, response)
+
         except requests.exceptions.ConnectionError as e:
             logging.exception(
                 'There was a connection error, check your network settings.', e)
@@ -162,7 +198,7 @@ class Google:
     @staticmethod
     def find_app_id(inputString):
         try:
-            searchResult = re.search('"appid":"[a-z://.-_]+"', inputString).group()
+            searchResult = re.search('"appid":"[a-z://.-_] + "', inputString).group()
             searchObject = json.loads('{' + searchResult + '}')
             return str(searchObject['appid'])
         except:
@@ -171,50 +207,37 @@ class Google:
 
     def do_login(self):
         self.session = requests.Session()
-        self.session.headers['User-Agent'] = "AWS Sign-in/{} (aws-google-auth)".format(self.version)
+        # self.session.headers['User-Agent'] = "AWS Sign-in/{} (aws-google-auth)".format(self.version)
+        self.session.headers['User-Agent'] = "aws-cli/1.16.215 Python/3.7.3 Linux/4.14.133-113.105.amzn2.x86_64 botocore/1.12.205"
         sess = self.get(self.login_url)
 
         # Collect information from the page source
         first_page = BeautifulSoup(sess.text, 'html.parser')
-        gxf = first_page.find('input', {'name': 'gxf'}).get('value')
+        # gxf = first_page.find('input', {'name': 'gxf'}).get('value')
         self.cont = first_page.find('input', {'name': 'continue'}).get('value')
-        page = first_page.find('input', {'name': 'Page'}).get('value')
-        sign_in = first_page.find('input', {'name': 'signIn'}).get('value')
-        account_login_url = first_page.find('form', {'id': 'gaia_loginform'}).get('action')
+        # page = first_page.find('input', {'name': 'Page'}).get('value')
+        # sign_in = first_page.find('input', {'name': 'signIn'}).get('value')
+        form = first_page.find('form', {'id': 'gaia_loginform'})
+        account_login_url = form.get('action')
 
-        payload = {
-            'bgresponse': 'js_disabled',
-            'checkConnection': '',
-            'checkedDomains': 'youtube',
-            'continue': self.cont,
-            'Email': self.config.username,
-            'gxf': gxf,
-            'identifier-captcha-input': '',
-            'identifiertoken': '',
-            'identifiertoken_audio': '',
-            'ltmpl': 'popup',
-            'oauth': 1,
-            'Page': page,
-            'Passwd': '',
-            'PersistentCookie': 'yes',
-            'ProfileInformation': '',
-            'pstMsg': 0,
-            'sarp': 1,
-            'scc': 1,
-            'SessionState': '',
-            'signIn': sign_in,
-            '_utf8': '?',
-        }
+        payload = {}
+
+        for tag in form.find_all('input'):
+            if tag.get('name') is None:
+                continue
+
+            payload[tag.get('name')] = tag.get('value')
+
+        payload['Email'] = self.config.username
 
         if self.config.bg_response:
             payload['bgresponse'] = self.config.bg_response
 
-        # GALX is sometimes not there
-        try:
-            galx = first_page.find('input', {'name': 'GALX'}).get('value')
-            payload['GALX'] = galx
-        except:
-            pass
+        if payload.get('PersistentCookie', None) is not None:
+            payload['PersistentCookie'] = 'yes'
+
+        if payload.get('TrustDevice', None) is not None:
+            payload['TrustDevice'] = 'on'
 
         # POST to account login info page, to collect profile and session info
         sess = self.post(account_login_url, data=payload)
@@ -306,6 +329,8 @@ class Google:
             sess = self.handle_sk(sess)
         elif "challenge/iap/" in sess.url:
             sess = self.handle_iap(sess)
+        elif "challenge/dp/" in sess.url:
+            sess = self.handle_dp(sess)
         elif "challenge/ootp/5" in sess.url:
             raise NotImplementedError(
                 'Offline Google App OOTP not implemented')
@@ -325,7 +350,6 @@ class Google:
     def parse_saml(self):
         if self.session_state is None:
             self.do_login()
-            # raise RuntimeError('You must use do_login() before calling parse_saml()')
 
         parsed = BeautifulSoup(self.session_state.text, 'html.parser')
         try:
@@ -335,8 +359,8 @@ class Google:
             if self.save_failure:
                 logging.error("SAML lookup failed, storing failure page to "
                               "'saml.html' to assist with debugging.")
-                with open("saml.html", 'w') as out:
-                    out.write(str(self.session_state.text.encode('utf-8')))
+                with open("saml.html", 'wb') as out:
+                    out.write(self.session_state.text.encode('utf-8'))
 
             raise ExpectedGoogleException('Something went wrong - Could not find SAML response, check your credentials or use --save-failure-html to debug.')
 
@@ -378,18 +402,22 @@ class Google:
             if find_executable('xv') is None and find_executable('display') is None:
                 open_image = False
 
-        print("Please visit the following URL to view your CAPTCHA: {}".format(captcha_url))
+        self._out.print(f"\nPlease visit the following URL to view your CAPTCHA: {captcha_url}")
+        self._out.notify("\n\nTo avoid having to enter captchas in the future, check your gmail and authorize the new "
+                         "device.")
 
         if open_image:
             try:
                 with requests.get(captcha_url) as url:
-                    subprocess.call(['open', url.content])
-
+                    with io.BytesIO(url.content) as f:
+                        Image.open(f).show()
             except Exception:
                 pass
 
-
-        captcha_input = input("Captcha (case insensitive): ") or None
+        try:
+            captcha_input = raw_input("Captcha (case insensitive): ") or None
+        except NameError:
+            captcha_input = input("Captcha (case insensitive): ") or None
 
         # Update the payload
         payload['identifier-captcha-input'] = captcha_input
@@ -404,35 +432,18 @@ class Google:
         newPayload = {}
 
         auth_response_page = BeautifulSoup(response.text, 'html.parser')
+        form = auth_response_page.find('form')
+        for tag in form.find_all('input'):
+            if tag.get('name') is None:
+                continue
 
-        challengeId = auth_response_page.find('input', {
-            'name': 'challengeId'
-        }).get('value')
-        challengeType = auth_response_page.find('input', {
-            'name': 'challengeType'
-        }).get('value')
-        tl = auth_response_page.find('input', {
-            'name': 'TL'
-        }).get('value')
-        gxf = auth_response_page.find('input', {
-            'name': 'gxf'
-        }).get('value')
+            newPayload[tag.get('name')] = tag.get('value')
 
-        newPayload['challengeId'] = challengeId
-        newPayload['challengeType'] = challengeType
-        newPayload['TL'] = tl
-        newPayload['gxf'] = gxf
-        newPayload['TrustDevice'] = "on"
-        newPayload['continue'] = payload['continue']
-        newPayload['ltmpl'] = payload['ltmpl']
-        newPayload['scc'] = payload['scc']
-        newPayload['sarp'] = payload['sarp']
-        newPayload['checkedDomains'] = payload['checkedDomains']
-        newPayload['checkConnection'] = payload['checkConnection']
-        newPayload['pstMsg'] = payload['pstMsg']
-        newPayload['oauth'] = payload['oauth']
-        newPayload['Email'] = payload['Email']
-        newPayload['Passwd'] = payload['Passwd']
+        newPayload['Email'] = self.config.username
+        newPayload['Passwd'] = self.config.password
+
+        if newPayload.get('TrustDevice', None) is not None:
+            newPayload['TrustDevice'] = 'on'
 
         return self.post(response.url, data=newPayload)
 
@@ -532,48 +543,23 @@ class Google:
 
         sms_token = input("Enter SMS token: G-") or None
 
-        payload = {
-            'challengeId':
-            response_page.find('input', {
-                'name': 'challengeId'
-            }).get('value'),
-            'challengeType':
-            response_page.find('input', {
-                'name': 'challengeType'
-            }).get('value'),
-            'continue':
-            response_page.find('input', {
-                'name': 'continue'
-            }).get('value'),
-            'scc':
-            response_page.find('input', {
-                'name': 'scc'
-            }).get('value'),
-            'sarp':
-            response_page.find('input', {
-                'name': 'sarp'
-            }).get('value'),
-            'checkedDomains':
-            response_page.find('input', {
-                'name': 'checkedDomains'
-            }).get('value'),
-            'pstMsg':
-            response_page.find('input', {
-                'name': 'pstMsg'
-            }).get('value'),
-            'TL':
-            response_page.find('input', {
-                'name': 'TL'
-            }).get('value'),
-            'gxf':
-            response_page.find('input', {
-                'name': 'gxf'
-            }).get('value'),
-            'Pin':
-            sms_token,
-            'TrustDevice':
-            'on',
-        }
+        challenge_form = response_page.find('form')
+        payload = {}
+        for tag in challenge_form.find_all('input'):
+            if tag.get('name') is None:
+                continue
+
+            payload[tag.get('name')] = tag.get('value')
+
+        if response_page.find('input', {'name': 'TrustDevice'}) is not None:
+            payload['TrustDevice'] = 'on'
+
+        payload['Pin'] = sms_token
+
+        try:
+            del payload['SendMethod']
+        except KeyError:
+            pass
 
         # Submit IPP (SMS code)
         return self.post(challenge_url, data=payload)
@@ -596,7 +582,7 @@ class Google:
 
         self.check_prompt_code(response_page)
 
-        print("Open the Google App, and tap 'Yes' on the prompt to sign in ...")
+        self._out.print("Open the Google App, and tap 'Yes' on the prompt to sign in ...")
 
         self.session.headers['Referer'] = sess.url
 
@@ -604,7 +590,7 @@ class Google:
         response = None
         while retry:
             try:
-                response = self.post(await_url, json=await_body)
+                response = self.post(await_url, json_data=await_body)
                 retry = False
             except requests.exceptions.HTTPError as ex:
 
@@ -672,7 +658,7 @@ class Google:
         """
         num_code = response.find("div", {"jsname": "EKvSSd"})
         if num_code:
-            print("numerical code for prompt: {}".format(num_code.string))
+            self._out.print("numerical code for prompt: {}".format(num_code.string))
 
     def handle_totp(self, sess):
         response_page = BeautifulSoup(sess.text, 'html.parser')
@@ -710,6 +696,24 @@ class Google:
         # Submit TOTP
         return self.post(challenge_url, data=payload)
 
+    def handle_dp(self, sess):
+        response_page = BeautifulSoup(sess.text, 'html.parser')
+
+        input("Check your phone - after you have confirmed response press ENTER to continue.") or None
+
+        form = response_page.find('form', {'id': 'challenge'})
+        challenge_url = 'https://accounts.google.com' + form.get('action')
+
+        payload = {}
+        for tag in form.find_all('input'):
+            if tag.get('name') is None:
+                continue
+
+            payload[tag.get('name')] = tag.get('value')
+
+        # Submit Configuration
+        return self.post(challenge_url, data=payload)
+
     def handle_iap(self, sess):
         response_page = BeautifulSoup(sess.text, 'html.parser')
         challenge_url = sess.url.split("?")[0]
@@ -830,110 +834,49 @@ class Google:
 
     def handle_selectchallenge(self, sess):
         response_page = BeautifulSoup(sess.text, 'html.parser')
-        # Known mfa methods, 5 is disabled till its implemented
-        auth_methods = {
-            2: 'TOTP (Google Authenticator)',
-            3: 'SMS',
-            4: 'OOTP (Google Prompt)'
-            # 5: 'OOTP (Google App Offline Security Code)'
-        }
 
-        unavailable_challenge_ids = [
-            int(i.attrs.get('data-unavailable'))
-            for i in response_page.find_all(
-                lambda tag: tag.name == 'form' and 'data-unavailable' in tag.attrs
-            )
-        ]
+        challenges = []
+        for i in response_page.select('form[data-challengeentry]'):
+            action = i.attrs.get("action")
 
-        # ootp via google app offline code isn't implemented. make sure its not valid.
-        unavailable_challenge_ids.append(5)
+            if "challenge/totp/" in action:
+                challenges.append(['TOTP (Google Authenticator)', i.attrs.get("data-challengeentry")])
+            elif "challenge/ipp/" in action:
+                challenges.append(['SMS', i.attrs.get("data-challengeentry")])
+            elif "challenge/iap/" in action:
+                challenges.append(['SMS other phone', i.attrs.get("data-challengeentry")])
+            elif "challenge/sk/" in action:
+                challenges.append(['YubiKey', i.attrs.get("data-challengeentry")])
+            elif "challenge/az/" in action:
+                challenges.append(['Google Prompt', i.attrs.get("data-challengeentry")])
 
-        challenge_ids = [
-            int(i.get('value'))
-            for i in response_page.find_all('input', {'name': 'challengeId'})
-            if int(i.get('value')) not in unavailable_challenge_ids
-        ]
+        self._out.print('Choose MFA method from available:')
+        for i, mfa in enumerate(challenges, start=1):
+            self._out.print("{}: {}".format(i, mfa[0]))
 
-        challenge_ids.sort()
+        selected_challenge = input("Enter MFA choice number (1): ") or None
 
-        auth_methods = {
-            k: auth_methods[k]
-            for k in challenge_ids
-            if k in auth_methods and k not in unavailable_challenge_ids
-        }
-
-        print('Choose MFA method from available:')
-        print('\n'.join(
-            '{}: {}'.format(*i) for i in list(auth_methods.items())))
-
-        selected_challenge = input("Enter MFA choice number ({}): ".format(
-            challenge_ids[-1:][0])) or None
-
-        if selected_challenge is not None and int(selected_challenge) in challenge_ids:
-            challenge_id = int(selected_challenge)
+        if selected_challenge is not None and int(selected_challenge) <= len(challenges):
+            selected_challenge = int(selected_challenge) - 1
         else:
-            # use the highest index as that will default to prompt, then sms, then totp, etc.
-            challenge_id = challenge_ids[-1:][0]
+            selected_challenge = 0
 
-        print("MFA Type Chosen: {}".format(auth_methods[challenge_id]))
-
-        # if challenge_id == 2:
-        #     if self._defaults.mfa_enabled:
-        #         color = Utils.default_colors() if self._defaults.colors_enabled else None
-        #         mfa_token = self._secrets_mgr.get_next_mfa(self._defaults.user) if self._defaults.auto_mfa else \
-        #             Input.get_mfa(display_hint=True, color=color)
-        #     else:
-        #         mfa_token = None
+        challenge_id = challenges[selected_challenge][1]
+        self._out.print("MFA Type Chosen: {}".format(challenges[selected_challenge][0]))
 
         # We need the specific form of the challenge chosen
         challenge_form = response_page.find(
             'form', {'data-challengeentry': challenge_id})
 
-        payload = {
-            'challengeId':
-            challenge_id,
-            'challengeType':
-            challenge_form.find('input', {
-                'name': 'challengeType'
-            }).get('value'),
-            'continue':
-            challenge_form.find('input', {
-                'name': 'continue'
-            }).get('value'),
-            'scc':
-            challenge_form.find('input', {
-                'name': 'scc'
-            }).get('value'),
-            'sarp':
-            challenge_form.find('input', {
-                'name': 'sarp'
-            }).get('value'),
-            'checkedDomains':
-            challenge_form.find('input', {
-                'name': 'checkedDomains'
-            }).get('value'),
-            'pstMsg':
-            challenge_form.find('input', {
-                'name': 'pstMsg'
-            }).get('value'),
-            'TL':
-            challenge_form.find('input', {
-                'name': 'TL'
-            }).get('value'),
-            'gxf':
-            challenge_form.find('input', {
-                'name': 'gxf'
-            }).get('value'),
-            'subAction':
-            challenge_form.find('input', {
-                'name': 'subAction'
-            }).get('value'),
-        }
-        if challenge_form.find('input', {'name': 'SendMethod'}) is not None:
-            payload['SendMethod'] = challenge_form.find(
-                'input', {
-                    'name': 'SendMethod'
-                }).get('value')
+        payload = {}
+        for tag in challenge_form.find_all('input'):
+            if tag.get('name') is None:
+                continue
+
+            payload[tag.get('name')] = tag.get('value')
+
+        if response_page.find('input', {'name': 'TrustDevice'}) is not None:
+            payload['TrustDevice'] = 'on'
 
         # POST to google with the chosen challenge
         return self.post(
